clear;  close all;  clc;

% 1D test function
f = @stk_testfun_braninhoo;
DIM = 2;
BOX = stk_hrect ([[-5; 10] [0; 15]], {'x1' 'x2'});

% Space discretization
NT = 100;  % Number of points in the grid
xg = stk_sampling_regulargrid (NT^2, DIM, BOX);

% Values of the function on the grid
zg = stk_feval (f, xg);
zg.colnames = {'z'};

s_title = 'The Branin-Hoo function';
stk_figure (s_title);  pcolor (xg, zg);  colorbar ();


%% Initial design of experiments
%
% We start with an initial design of N0 points, regularly spaced on the domain.
%

% Size of the initial design
N0 = 10;  % 10 * DIM;

% Construction of the initial design
x0 = stk_sampling_maximinlhs (N0, DIM, BOX);

% Values of the function on the initial design
z0 = stk_feval (f, x0);
z0.colnames = {'z'};


%% Sequential design of experiments

% Start with the initial design defined above
x = x0;  z = z0;

% GP model
M_prior = stk_model ('stk_materncov52_aniso', DIM);

% Number of points to be added adaptively
NB_ITER = 40;

% Current value of the maximum of the Expected Improvement
EI_max = +inf;

% Value of EI_max for the stopping criterion
EI_max_stop = (max (zg) - min (zg)) / 1e7;

% Iteration number
iter = 0;

while (iter < NB_ITER) && (EI_max > EI_max_stop)
    
    % Trick: add a small "regularization" noise to our model
    M_prior.lognoisevariance = 2 * log (min (1e-4, EI_max / 1e3));
    
    % Compute posterior model
    % (variance and range parameters are estimated by ReML)
    M_post = stk_model_gpposterior (M_prior, x, z);

    % Carry out the kriging prediction
    z_pred = stk_predict (M_post, xg);
    
    % Compute the Expected Improvement (EI) criterion on the grid
    EI_val = stk_sampcrit_ei_eval (z_pred.mean, sqrt (z_pred.var), z);
    
    % Pick the point where the EI is maximum as our next evaluation point
    [EI_max, i_max] = max (EI_val);
    
    
    figure (12);
    
    subplot (2, 2, 1);  cla;  pcolor (xg, z_pred.mean);  hold on;
    plot (x(:, 1), x(:, 2), 'ro', 'MarkerFaceColor', 'r');
    plot (x(end, 1), x(end, 2), 'ro', 'MarkerFaceColor', 'y');
    stk_title ('posterior mean');
    
    % subplot (2, 2, 2);  cla;  pcolor (xg, log10 (z_pred.var));
    % stk_title ('posterior var (log10)');
    subplot (2, 2, 2);  cla;  pcolor (xg, EI_val);
    stk_title ('sampling crit');
    
    % Compye LOO predictions and residuals
    [z_LOO, res_LOO] = stk_predict_leaveoneout (M_post);
    subplot (2, 2, 3);  cla;  stk_plot_predvsobs (z, z_LOO);
    subplot (2, 2, 4);  cla;  stk_plot_histnormres (res_LOO.norm_res);
    
    if EI_max > EI_max_stop
        % Add the new evaluation to the DoE
        x = vertcat (x, xg(i_max, :));
        z = vertcat (z, zg(i_max, :));
        iter = iter + 1;
    end
    
    drawnow;  pause;
    
end

% Display the final DoE
disp (x);

% Total number of evaluations ?
fprintf ('\nNumber of evaluations: %d + %d = %d.\n\n', N0, iter, N0 + iter);


%#ok<*AGROW>

%!test stk_example_doe03;  close all;
