% STK_OPTIM_EHVI optimizes the EHVI criteria using a given algorithm
%
% CALL: [x_opt, ehvi_opt, AUX] = stk_optim_ehvi ...
%          (OPTIONS, DOM, MODEL, DATA, Z_REF, AUX)
%
%    optimizes the EHVI criteria with a scatter plot of POPULATIONSIZE points.
%    Optimization done given a MODEL and some DATA in the domain DOM, with
%    respect to a reference Z_REF.
%    Some points of DATA may not be evaluated yet. A realisation is simulated
%    instead, given the MODEL.
%
% AVAILABLE OPTIMIZATION METHODS
%
%    The following methods can be selected by setting OPTIONS.method
%    to the appropriate value:
%
%     * SRS: Simple (IID, uniform) random search, fast but not very efficient,
%
%     * SMC: Sequential Monte Carlo as in [1, 2], recommended as a default,
%
%     * GA: Matlab genetic algorithm (from the Global Optimization toolbox).
%
% REFERENCES
%
%   [1] Romain Benassi.  Nouvel algorithme d'optimisation bayesien
%       utilisant une approche Monte-Carlo sequentielle.  PhD Thesis,
%       Ecole Superieure d'Electricite (Supelec), 2013.
%
%   [2] Paul Feliot, Julien Bect, Emmanuel Vazquez.  A Bayesian approach
%       to constrained single- and multi-objective optimization.
%       Journal of Global Optimization, 67(1):97-133, 2017
%
% See also stk_optim_ehvi_grid, stk_optim_ehvi_smc, stk_example_boo01,
%          stk_example_boo02, stk_example_boo03

% Copyright Notice
%
%    Copyright (C) 2017 CentraleSupelec
%
%    Authors:  Etienne Leloup           <etienne.leloup@free.fr>
%              Guillaume Maistre-Bazin  <g.maistre.bazin@gmail.com>
%              Lucain Pouget            <lucainp@gmail.com>
%              Julien Bect              <julien.bect@centralesupelec.fr>

% Copying Permission Statement
%
%    This file is part of STK-contrib-moo-1
%      (https://sourceforge.net/p/kriging/contrib-moo-1/).
%
%    STK-contrib-moo-1 is free software;  you can redistribute it  and/or
%    modify it  under the terms of the  GNU Lesser General Public License
%    as published by the Free Software Foundation;  either version 2.1 of
%    the License, or (at your option) any later version.
%
%    STK-contrib-moo-1 is distributed in the hope that it will be useful,
%    but  WITHOUT  ANY  WARRANTY;  without even  the implied  warranty of
%    MERCHANTABILITY or FITNESS  FOR A  PARTICULAR PURPOSE.  See the  GNU
%    Lesser General Public License for more details.
%
%    You should  have received  a copy  of the  GNU Lesser General Public
%    License along with STK-contrib-moo-1;  if not, see
%                                          <http://www.gnu.org/licenses/>.

function [x_out, ehvi_out, aux] = stk_optim_ehvi ...
    (options, dom, model, data, z_ref, aux)

if nargin < 6
    aux = [];
end

% Default batch size
if ~ isfield (options, 'batch_size') || isempty (options.batch_size)
    options.batch_size = 1;
end

bs = options.batch_size;

% Default global optimizer
% FIXME: Rename 'method' -> 'global_optimizer'
if ~ isfield (options, 'method') || isempty (options.method)
    options.method = 'SMC';
end

% Default local optimizer
if ~ isfield (options, 'local_optimizer') || isempty (options.local_optimizer)
    options.local_optimizer = 'fmincon';
end

% FIXME: change input argument 'data' to 'x, z'
x = data(:, 1:end-2);
z = data(:, end-1:end);

for j = 1:bs
    
    if j > 1
        % Replace unavailable responses with random simulations 
        % (this is implicitely doing a very rough form of Monte Carlo
        %  estimation of the multi-point sampling criterion)
        data = simulate_nan_data (data, model);
    end
    
    [x_, ehvi_, aux_] = stk_optim_ehvi_ ...
        (options, dom, model, horzcat (x, z), z_ref, aux);
    
    % *** This is only relevant for the SMC case ***
    % After the selection of the first point in the batch,
    % restart each time from the same SMC sample
    if j == 1
        aux = aux_;
    end
    
    if j < bs
        
        % Fake evaluation
        % (while the point has not been evaluated, value is set to NaN)
        x_ = double (x_);
        z_ = nan (1, 2);
        
        % Add new (unevaluated) point
        x = vertcat (x, x_);
        z = vertcat (z, z_);
        
    end
end

if bs == 1
    x_out = x_;
    ehvi_out = ehvi_;
else
    x_out = x(end-bs+1:end, :);
    ehvi_out = nan;    
end

end % function

%#ok<*AGROW>

    
function [x_opt, ehvi_opt, aux] = stk_optim_ehvi_ ...
    (options, BOX, model, data, z_ref, aux)

switch options.method
    
    case {'SRS', 'IID'}  % Simple random search using a uniform IID sample
        
        [x_opt, ehvi_opt] = stk_optim_ehvi_srs ...
            (options, BOX, model, data, z_ref);
        
        aux = [];
        
    case 'GA'  % gamultiobj
        
        [x_opt, ehvi_opt] = stk_optim_ehvi_ga ...
            (options, BOX, model, data, z_ref);
        
        aux = [];
        
    case 'SMC' % SMC optimization
        
        [x_opt, ehvi_opt, aux] = stk_optim_ehvi_smc ...
            (options, BOX, model, data, z_ref, aux);
        
    otherwise
        stk_error ('Optimization method not recognized', 'InvalidArgument');
end

switch options.local_optimizer
    
    case 'fmincon'  % A few iterations of fmincon
        
        f = @(x) - compute_ehvi (data, x, model, z_ref);
        
        options = optimset ('fmincon');
        options.Display = 'none';
        options.MaxIter = 5;
        
        [x_opt, ehvi_opt] = fmincon (f, double (x_opt), ...
            [], [], [], [], BOX.lower_bounds, BOX.upper_bounds, [], options);
        
    case 'none'
        
        % Well, nothing to do then
        
    otherwise
        
        error ('Incorrect value for options.local_optimizer');
        
end

end % function
