% STK_SMC_INITIALIZE initializes the SMC algorithm with a uniform sample
%
% CALL: SMC = stk_smc_initialize (POPSIZE, DOMAIN)
%
%    creates a structure SMC representing sample of size POPSIZE distributed
%    uniformly on the hyper-rectangular DOMAIN.
%
% See also stk_optim_ehvi_smc, stk_smc_reweight,
%          stk_smc_resample, stk_smc_move

% Copyright Notice
%
%    Copyright (C) 2017 CentraleSupelec
%
%    Authors:  Julien Bect  <julien.bect@centralesupelec.fr>

% Copying Permission Statement
%
%    This file is part of STK-contrib-moo-1
%      (https://sourceforge.net/p/kriging/contrib-moo-1/).
%
%    STK-contrib-moo-1 is free software;  you can redistribute it  and/or
%    modify it  under the terms of the  GNU Lesser General Public License
%    as published by the Free Software Foundation;  either version 2.1 of
%    the License, or (at your option) any later version.
%
%    STK-contrib-moo-1 is distributed in the hope that it will be useful,
%    but  WITHOUT  ANY  WARRANTY;  without even  the implied  warranty of
%    MERCHANTABILITY or FITNESS  FOR A  PARTICULAR PURPOSE.  See the  GNU
%    Lesser General Public License for more details.
%
%    You should  have received  a copy  of the  GNU Lesser General Public
%    License along with STK-contrib-moo-1;  if not, see
%                                          <http://www.gnu.org/licenses/>.

function [smc, sigma] = stk_smc_initialize (pop_size, domain, updf_, data)

% Draw a uniform sample in the hyper-rectangle
x = stk_sampling_randunif (pop_size, [], domain);

% Make sure we have a plain double precision array
x = double (x);

% Store everything in a structre
smc.particles = x;
smc.weights   = repmat (1/pop_size, pop_size, 1);
smc.updf_fun  = @(x) ones (size (x, 1), 1);
smc.updf_val  = ones (pop_size, 1);

% Define initial sigma (the scaling in 1 / sqrt (dim) is loosely motivated by
% theoretical MCMC results, but the constant is rather arbitrary...)
dim = size (domain, 2);
sigma = (diff (domain)) / (sqrt (dim));

% Transition from uniform to target initial distribution
if nargin > 2 && ~ isempty (data)
    
    % Random permutation of evaluation results
    n_eval = size (data, 1);
    data = data(randperm (n_eval), :);
    
    % Absorb evaluations one by one (as in Chopin's 2002 paper)
    % FIXME: this is slow => absorb batches of evaluations
    for n = 1:n_eval
        stk_disp_progress ('stk_smc_initialize: %d/%d', n, n_eval);
        updf = @(x) updf_ (x, data(1:n, :));
        smc = stk_smc_reweight (smc, updf);  % FIXME: restart ???
        smc = stk_smc_resample (smc);
        [smc, sigma] = stk_smc_move (smc, sigma);
    end
end

end % function
