% STK_EXAMPLE_BOO02  An illustration of Bayesian bi-objective
% optimization. Example has multiple problems and parameters to play with
%
% Our goal here is to optimize a bi-objective problem. You can choose it :
%
%  Problem 1 :
%    (x1,x2) |--> x1^2 + (x2-7)^2
%    (x1,x2) |--> stk_testfun_braninhoo(x1,x2)
%    over [-5; 10] x [0; 15].
%
%  Problem 2 :
%    Unconstrained BNH problem. Please see reference for details [1]
%
%  Problem 3 ;
%    Schaffer problem. Please see reference for details [2]
%
%
% REFERENCES
%   [1] BV Babu, AM Gujarathi "Multi-objective differential evolution (mode)
%       algorithm for multi-objective optimization: parametric study on
%       benchmark test problems", 2007
%
%   [2] M. Emmerich and K. Giannakoglou and B. Naujoks "Single- and Multi-
%       objective Evolutionary Optimization Assisted by Gaussian Random
%       Field Metamodels"

% Copyright Notice
%
%    Copyright (C) 2017 CentraleSupelec
%
%    Authors:  Etienne Leloup           <etienne.leloup@free.fr>
%              Guillaume Maistre-Bazin  <g.maistre.bazin@gmail.com>
%              Lucain Pouget            <lucainp@gmail.com>
%              Julien Bect              <julien.bect@centralesupelec.fr>

% Copying Permission Statement
%
%    This file is part of STK-contrib-moo-1
%      (https://sourceforge.net/p/kriging/contrib-moo-1/).
%
%    STK-contrib-moo-1 is free software;  you can redistribute it  and/or
%    modify it  under the terms of the  GNU Lesser General Public License
%    as published by the Free Software Foundation;  either version 2.1 of
%    the License, or (at your option) any later version.
%
%    STK-contrib-moo-1 is distributed in the hope that it will be useful,
%    but  WITHOUT  ANY  WARRANTY;  without even  the implied  warranty of
%    MERCHANTABILITY or FITNESS  FOR A  PARTICULAR PURPOSE.  See the  GNU
%    Lesser General Public License for more details.
%
%    You should  have received  a copy  of the  GNU Lesser General Public
%    License along with STK-contrib-moo-1;  if not, see
%                                          <http://www.gnu.org/licenses/>.

%% Play with parameters to understand the cases

NIter = 30;       % Number of points to be added adaptively
NIterPause = Inf; % Number of points to be added between each pause
NEstimParam = 10; % Number of points to be added between each reestimation of the parameters

% Select the bi-objective problem
%  - 1 : polynomial + Brannin
%  - 2 : BNH
%  - 3 : Schaffer problem dim2
problem = 1;


%% Problem definition

switch problem
    
    case 1 % Polynomial + Brannin
        
        % Objective functions
        f1 = @(x) x(:,1) .^ 2 + (x(:,2) - 7) .^ 2;
        f2 = @(x) stk_testfun_braninhoo (x);
        
        % Input space (hyper-rectangle object)
        dom = stk_hrect ([ [-5; 10] [0; 15] ], {'x1', 'x2'});
        
        % Reference point in objective space
        z_ref = [200 200];
        
    case 2 % BNH
        
        % Objective functions
        f1 = @(x) 4 * x(:,1) .^ 2 + 4 * x(:,2) .^ 2;
        f2 = @(x) (x(:,1) - 5) .^ 2 + (x(:,2) - 5) .^ 2;
        
        % Input space (hyper-rectangle object)
        dom = stk_hrect ([[0; 5] [0; 3]], {'x1', 'x2'});
        
        % Reference point in objective space
        z_ref = [200 200];
        
    case 3 % Schaffer roblem 2D
        
        % gamma can be tuned to alter the shape of the Pareto front
        gamma = 1;
        cte = 1 / (2 ^ (gamma / 2));
        
        % Objective functions
        f1 = @(x) cte * (sum (x .* x, 2)) .^ (gamma / 2);
        f2 = @(x) cte * (sum ((1 - x) .^ 2, 2)) .^ (gamma / 2);
        
        % Input space (hyper-rectangle object)
        dom = stk_hrect ([[0; 10] [0; 10]], {'x1', 'x2'});
        
        % Reference point in objective space
        z_ref = [10 10] .^ gamma;
        
end

% Upper constraints on objective functions
% (setting them to +Inf means 'no constraint')
z_constr = [Inf Inf];

% Setting upper constraints is actually equivalent to changing the ref point
% (from now on, any point that does not dominate z_ref is considered unfeasible)
z_ref = min (z_ref, z_constr);


%% Initial design of experiments

% Size of the initial design
n0 = 20;

% Construction of the initial design
x0 = stk_sampling_maximinlhs (n0, [], dom);

% Values of the function on the initial design
z0 = stk_feval ({f1 f2}, x0);
z0.colnames = {'z1', 'z2'};


%% Display problem

figure ();  stk_plot_boo_2D_problem (dom, f1, f2, z_ref)


%% Specification of the model (Gaussian process prior)

dim = size (dom, 2);  % Dimension of the input space

model(1) = stk_model ('stk_materncov52_aniso', dim);
model(2) = stk_model ('stk_materncov52_aniso', dim);

% Estimate the parameters of the models based on initial design
model(1).param = stk_param_estim (model(1), x0, z0(:, 1));
model(2).param = stk_param_estim (model(2), x0, z0(:, 2));


%% Sequential design of experiments
%
% Here, evaluations points are chosen sequentially using the Expected
% Hypervolume Improvement (EHVI) criterion, starting from the initial
% design defined above.
%
% The algorithm stops when the maximum number of iterations is reached.

aux = [];
EHVI_max = +inf;
DHV_seq = zeros(NIter,1);

x = x0;
z = z0;

optim_options = struct ();
optim_options.method = 'SMC';        % See help stk_optim_ehvi for details
optim_options.population_size = [];  % empty means: use the default

for i = 1:NIter
    
    % Optimise EHVI using SMC
    [x_new, ehvi_opt, aux] = stk_optim_ehvi ...
        (optim_options, dom, model, horzcat (x, z), z_ref, aux);
    
    % Evaluation
    x_new = double (x_new);
    z_new = stk_feval ({f1 f2}, x_new);
    
    % Add new point
    x = vertcat (x, x_new);
    z = vertcat (z, z_new);
    
    % Extract non-dominated solutions
    idx_pareto = stk_paretofind_constrained (z, z_ref);
    x_pareto = x(idx_pareto, :);
    z_pareto = z(idx_pareto, :);
    
    % Compute dominated hypervolume
    DHV_seq(i) = stk_dominatedhv (z_pareto, z_ref);
    
    % Display progress if needed
    stk_plot_boo_2D (dom, model, x, z, z_ref, DHV_seq(1:i));  drawnow
    
    % Pause if needed
    if(mod(i, NIterPause) == 0)
        pause
    end
    
    % Estimation of parameters if needed
    if mod(i, NEstimParam) == 0
        model(1).param = stk_param_estim (model(1), x, z(:, 1));
        model(2).param = stk_param_estim (model(2), x, z(:, 2));
    end
end


%#ok<*AGROW>
